/*----------------------------------------------------------------------
  File    : solve.c
  Contents: solve linear equation systems
  Author  : Christian Borgelt
  History : 2001.10.17 file created
            2003.08.16 slight changes in error message output
            2007.02.13 adapted to modified module tabread
            2010.10.11 adapted to modified module tabread
            2010.12.11 adapted to a generic error reporting function
            2013.08.09 adapted to higher compiler warning level
            2013.08.13 adapted to definition of type DIMID
            2014.10.24 changed from LGPL license to MIT license
            2020.03.28 regularization parameters added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include "escape.h"
#include "tabread.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "solve"
#define DESCRIPTION "solve linear equation system"
#define VERSION     "version 2.2 (2020.03.28)         " \
                    "(c) 2001-2020   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in matrix.h */
#define E_STDIN     (-5)        /* double assignment of stdin */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_EQCNT     (-9)        /* number of unknowns/equations */
#define E_ALGO     (-10)        /* unknown algorithm */
#define E_SING     (-11)        /* singular coefficient matrix */
#define E_POSDEF   (-12)        /* matrix is not positive definite */
#define E_REGPARM  (-13)        /* invalid regularization params. */
/* error codes  -16 to -18 defined in matrix.h */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_EQCNT    -9 */  "number of unknowns and "
                         "number of equations differ",
  /* E_ALGO    -10 */  "unknown algorithm '%c'",
  /* E_SING    -11 */  "singular coefficient matrix",
  /* E_POSDEF  -12 */  "coefficient matrix is not positive definite",
  /* E_REGPARM -13 */  "invalid regularization parameters '%s'",
  /*    -14 to -15 */  NULL, NULL,
  /* E_VALUE   -16 */  "%s:%d(%d): invalid value '%s'",
  /* E_FLDCNT  -17 */  "%s:%d(%d): wrong number of fields/columns",
  /* E_RECCNT  -18 */  "%s:%d(%d): wrong number of records/rows",
  /*           -19 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static TABREAD *tread = NULL;   /* table/matrix reader */
static double  *vec   = NULL;   /* vector to read equations */
static double  *rhs   = NULL;   /* vector of right hand sides */
static MATRIX  *mat   = NULL;   /* matrix of coefficients */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (rhs)   free(vec); \
  if (vec)   free(vec); \
  if (mat)   mat_delete(mat); \
  if (tread) trd_delete(tread, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s, *r;               /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_les  = NULL;      /* linear equation system file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *fmt     = "%g";      /* format for number output */
  int     algo     = 'g';       /* algorithm to use */
  int     mode     = MAT_PARTPIV;  /* pivoting mode */
  int     cond     = 0;         /* flag for condensed output */
  CCHAR   *regparm = NULL;      /* regularization parameters */
  double  reg      =  0.0;      /* standard regularization parameter */
  double  rmin     =  0.0;      /* minimum  regularization parameter */
  double  rmax     = -1.0;      /* maximum  regularization parameter */
  DIMID   cnt      = 0;         /* number of equations/unknows */
  char    seps[4]  = " \n";     /* separators for output */
  DIMID   n;                    /* number of rows/equations */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] lesfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-a#      algorithm to use                  "
                    "(default: %c)\n", algo);
    printf("         g: Gauss-Jordan elimination "
                    "(any regular matrix)\n");
    printf("         l: LU       decomposition   "
                    "(any regular matrix)\n");
    printf("         c: Cholesky decomposition   "
                    "(symmetric positive definite matrix)\n");
    printf("-p       full pivoting (Gauss-Jordan only)   "
                    "(default: partial pivoting)\n");
    printf("-t#      Tikhonov regularization parameters  "
                    "(default: %g:%g:%g)\n", reg, rmin, rmax);
    printf("         (Cholesky decomposition only)\n");
    printf("-o#      format for number output            "
                    "(default: \"%s\")\n", fmt);
    printf("-c       condensed output "
                    "(print only a vector of numbers)\n");
    printf("-r#      record  separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                  "
                    "(default: \"#\")\n");
    printf("lesfile  file to read linear equation system from\n"
           "         (numbers only, right hand side in last field)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'a': algo   = (*s) ? *s++ : 'g';     break;
          case 'p': mode   = MAT_FULLPIV;           break;
          case 't': optarg = &regparm;              break;
          case 'o': optarg = &fmt;                  break;
          case 'c': cond   = 1;                     break;
          case 'r': optarg = &recseps;              break;
          case 'f': optarg = &fldseps;              break;
          case 'b': optarg = &blanks;               break;
          case 'C': optarg = &comment;              break;
          default : error(E_OPTION, *--s);          break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_les = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (!strchr("glc012", algo)) error(E_ALGO, algo);
  while (((algo == 'c') || (algo == '2')) && regparm) {
    reg = strtod(r = s, &s);    /* if regularizaton params. given */
    if (s == r) error(E_REGPARM, regparm);
    if (*s != ':') break;       /* get standard regularization parm. */
    rmin = strtod(r = s+1, &s);
    if (s == r) error(E_REGPARM, regparm);
    if (*s != ':') break;       /* get minimum regularization parm. */
    rmax = strtod(r = s+1, &s);
    if (*s || (s == r)) error(E_REGPARM, regparm);
    break;                      /* get maximum regularization parm. */
  }
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read the equation system --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_les) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  k = vec_readx(&vec, &cnt, tread);  /* read the first vector */
  if (k) error(k, TRD_INFO(tread));  /* and check for a read error */
  if (cnt < 2) error(E_FLDCNT, TRD_INFO(tread));
  rhs = (double*)malloc((size_t)--cnt *sizeof(double));
  mat = mat_create(cnt, cnt);   /* create the coefficient matrix and */
  if (!mat || !rhs) error(E_NOMEM);  /* a vector of right hand sides */
  n = 0;                        /* initialize the equation counter */
  do {                          /* and read the equations */
    if (n >= cnt) error(E_EQCNT);
    mat_rowset(mat, n, vec);    /* copy coefficients to the matrix */
    rhs[n++] = vec[cnt];        /* and set corresp. right hand side */
  } while ((k = vec_read(vec, cnt+1, tread)) == 0);
  if (k < 0) error(k, TRD_INFO(tread));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  if (n != cnt) error(E_EQCNT); /* check the number of equations */
  fprintf(stderr, "[%"DIMID_FMT" equations(s)]", cnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- solve the equation system --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "solving linear equation system ... ");
  switch (algo) {               /* evaluate the algorithm code */
    case 'g': case '0':         /* Gauss-Jordan elimination */
      if (mat_gjsol(&rhs, mat, &rhs, 1, mode, NULL) != 0)
      { error(E_SING); }          break;
    case 'l': case '1':         /* LU decomposition */
      if (mat_ludecom(mat, mat) != 0) error(E_SING);
      mat_lusubst(rhs, mat, rhs); break;
    case 'c': case '2':         /* Cholesky decomposition */
      if (mat_chdecx(mat, mat, reg, rmin, rmax) != 0) error(E_POSDEF);
      mat_chsubst(rhs, mat, rhs); break;
    default: error(E_ALGO, algo);
  }                             /* (solution is in vector rhs) */
  fprintf(stderr, "[%"DIMID_FMT" variable(s)]", cnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- print the result --- */
  if (cond) {                   /* if condensed output */
    if (fldseps) seps[0] = (char)esc_decode(fldseps, NULL);
    if (recseps) seps[1] = (char)esc_decode(recseps, NULL);
    for (n = 0; n < cnt; n++) { /* traverse the vector elements */
      if (n > 0) putchar(seps[0]);
      printf(fmt, rhs[n]);      /* print a separator and */
    }                           /* the next vector element */
    putchar(seps[1]); }         /* terminate the output record */
  else {                        /* if normal/verbose output, */
    for (n = 0; ++n <= cnt; ) { /* traverse the result vector */
      printf("x_%"DIMID_FMT" = ", n);
      printf(fmt, rhs[n-1]); printf("\n");
    }                           /* print variable/value pairs */
  }

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
